#ifndef CXXTLS_LEXICAL_CAST_H_INCLUDED_P
#define CXXTLS_LEXICAL_CAST_H_INCLUDED_P
//
// Copyright 2012, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

///@file
///  This file defines tools that allow you to cast objects from one form to
///  another where the objects don't technically cast in a C++ standard
///  way.  See cxxtls::lexical_cast.

#include <cxxtls/fmtspec.h>
#include <cxxtls/classTraits.h>
#include <portable_io.h>

#include <string>
#include <ctype.h>
#include <limits.h>


namespace cxxtls
{

    template<class To, class From>
    struct bad_lexical_cast
        /// There are numerous ways that this exception type can be thrown.  Not worth documenting.
        /// If you get this exception, just track down your mistake because, as Hermione says:  "You're
        /// doing it WRONG!
    {
    };


    template<typename T, typename U>
    struct trivial_cast
       /// A trivial_cast object has no member data but it does have a static
       /// function, castFunc(), that will copy an input parameter of type U
       /// as into value of type T.
       ///
       /// This class exists only so that it can used by lexical_cast_helper in
       /// the case where a trivial cast operation is allowed.
    {
        static void castFunc(T &output, U input)   // trivial cast
        {
           output = input;
        }
    };

    template<class T, class F>
    struct You_Forgot_To_Overload_cxxtls_lexical_cast_With_The_Following_Parms // no cast is possible
        /// A class that is only used when you have done something wrong.  It helps you get
        /// more intelligent error messages
    {
       static void castFunc(T, F);  // no implementation

       /// If you get a link error on this class it is because you attempted to perform a
       /// cxxtls::lexical_cast between two types for which no conversion is defined.  If you need
       /// for lexical_cast to work for the types involved, you must define an overload
       /// for the cxxtls::lexical_cast function that makes it work.
       ///
       /// Do not define a body for this class/function...  define lexical_cast<> instead
    };

    template<class F>
    struct signed_to_string_cast
        /// A class that wraps the cast function that converts signed integers to
        /// strings.  See castFunc().
        ///
        /// The parameter F represents the "from type" of a conversion to string.
        /// F is assumed to be a signed integer:  short, int, long, or long long.
    {
        static void castFunc(std::string& rv, F const &p)
            /// convert to string from some sort of signed integer using the default
            /// conversion format.
        {
            char buffer[40];
            signedInt2String(buffer, p, 'd', fmtspec());
            buffer[39]=0; // just in case

            rv = buffer;
        }

        static void castFunc(std::string& rv, F const &p, fmtspec s)
            /// convert to string from some sort of signed integer using a user
            /// defined format.
        {
            char buffer[40];
            signedInt2String(buffer, p, 'd', s);
            buffer[39]=0; // just in case

            rv = buffer;
        }

    };

    template<class F>
    struct float_to_string_cast
        /// A class that wraps the cast function that converts floats to
        /// strings.  See castFunc().
        ///
        /// The parameter F represents the "from type" of a conversion to string.
        /// F is assumed to be a 'float' or 'double'.  I wouldn't know what to do
        /// with long double...
    {
        static void castFunc(std::string& rv, F const &p)
           /// convert a float to a string using the default format (g)
        {
           double tmp = p;

           char buffer[40];

           snprintf(buffer, sizeof(buffer), "%lg", tmp);
           buffer[39]=0; // just in case

           rv = buffer;

        }

        static void castFunc(std::string& rv, F const &p, fmtspec s)
            /// convert a float to a string using a user defined
            /// format specification -- up to 400 bytes.
        {
           double tmp = p;

           char buffer[400];
           char cs[40];

           s.preferredStyle('g');

           snprintf(buffer, sizeof(buffer), s.cs(cs,1), tmp);
           buffer[399]=0;

           rv = buffer;

        }

    };

    template<class F>
    struct unsigned_to_string_cast
        /// A class that wraps the cast function that converts unsigned integers to
        /// strings.  See castFunc().
        ///
        /// The parameter F represents the "from type" of a conversion to string.
        /// F is assumed to be a unsigned integer.
    {
        static void castFunc(std::string &rv, F const &p)
            /// convert to string from some sort of unsigned integer
            /// using the default conversion format.
        {
            char buffer[40];
            signedInt2String(buffer, p, 'u', fmtspec());
            buffer[39]=0;

            rv = buffer;
        }

        static void castFunc(std::string &rv, F const &p, fmtspec s)
            /// convert to string from some sort of unsigned integer
            /// using a user defined format specification.
        {
            char buffer[40];
            signedInt2String(buffer, p, 'u', s);
            buffer[39]=0;

            rv = buffer;
        }

    };


    template<class T>
    struct string_to_signed_cast
        /// A class that wraps the cast function that converts std::string to
        /// signed integers.  See castFunc().
        ///
        /// The parameter T represents the "To type" of a conversion from std::string.
        /// T is assumed to be a signed integer.
    {
        static void castFunc(T& rv, std::string const &s)
            /// convert from string to some sort of signed integer
        {
            long long tmp=0;

            if(1 > sscanf(s.c_str(), "%lld", &tmp) )
            {
                throw bad_lexical_cast<T, std::string>();
            }

            rv = tmp;

        }

    };

    template<class T>
    struct string_to_float_cast
        /// A class that wraps the cast function that converts std::string to
        /// signed floats.  See castFunc().
        ///
        /// The parameter T represents the "To type" of a conversion from std::string.
        /// T is assumed to be a float.
    {
        static void castFunc(T& rv, std::string const &s)
            /// convert from string to some sort of signed integer
        {
            double tmp=0;

            if(1 > sscanf(s.c_str(), "%lf", &tmp) )
            {
                throw bad_lexical_cast<T, std::string>();
            }

            rv = tmp;

        }
    };

    template<class T>
    struct string_to_unsigned_cast
        /// A class that wraps the cast function that convers std::string to
        /// unsigned integers.  See castFunc().
        ///
        /// The parameter T represents the "To type" of a conversion from std::string.
        /// T is assumed to be a unsigned integer.
    {
        static void castFunc(T& rv, std::string const &s)
            /// convert from string to some sort of unsigned integer
        {
            unsigned long long tmp=0;

            if(1 > sscanf(s.c_str(), "%llu", &tmp) )
            {
                throw bad_lexical_cast<T, std::string>();
            }

            rv = tmp;
        }
    };

    template<class T, class F>
    struct lexical_cast_helper
        /// Define the nested type, 'type', to refer to a class which has a member
        /// function signature, castFunc(), that will let us perform casts and actual
        /// conversions as needed -- if possible.  The only casts or conversions that can
        /// be performed are those which are specifically programmed either with 
        /// tricks used by this class or by the user defining an overload for the
        /// lexical_cast signature.  Said signature would preclude this class from
        /// ever beig involved.
    {
        // To understand what is going on in this class, look for the definition of the
        // nested type, 'type', and work your way backwards.  Basically, there is a recursive
        // binary decomposition of possible interpretations of the meaing of cast from F to T.
        // Each possible meaning is imbedded in a type which is, in itself, recursively decomposed
        // into possible meanings.

        // Disambiguate the kind of cast that we are going to perform by selecting an
        // execution type based on the nature of the parameters.

        typedef typename removeReference<F>::type  nonRefF;  
             // remove the reference characterstic from type attributes of F.

        typedef typename removeCV< nonRefF >::type cleanF;   
            // same as F but not const, volatile, or reference


        // the following constants let us know the basic attributes of the data types
        // presented to us as F and T.  

        enum c1 { isTrivialConversion = isConvertibleType<F,T>::value,

                  // Is type F a string??

                  inputIsString       =     isSameType<cleanF, std::string>::value 
                                        ||  isSameType<cleanF, char *>::value
                                        ||  isSameType<cleanF, const char *>::value,


                  // Is type T a number?

                  outputIsString      =     isSameType<T, std::string>::value 
                                        ||  isSameType<T, char *>::value
                                        ||  isSameType<T, char const *>::value,

                  // ditoo and reverse

                  outputIsNumber      =     isIntegralType<T>::value 
                                        ||  isFloatType<T>::value,


                  inputIsNumber       =     isIntegralType<cleanF>::value
                                        ||  isFloatType<cleanF>::value,

                };

        typedef typename EvalTypeIf<
                                           isFloatType<F>::value,
                                           signed_to_string_cast<F>, // integers
                                           float_to_string_cast<F>
                                   >::type
                                   signed_to_string_conversion;
                                      // F is either a signed integer or a float
                                      // and T is a std::string 



        typedef typename EvalTypeIf<
                                            isUnsignedType<cleanF>::value,
                                            signed_to_string_conversion,
                                            unsigned_to_string_cast<F>
                                   >::type
                                   number_to_string_conversion;
                                      // F is a nubmer and T is a string


        typedef typename EvalTypeIf<        isFloatType<T>::value,
                                            string_to_signed_cast<T>,
                                            string_to_float_cast<T>
                                   >::type
                                   string_to_signed;
                                     // F is string and T is signed integer or float

        typedef typename EvalTypeIf<
                                            isUnsignedType<T>::value,
                                            string_to_signed,
                                            string_to_unsigned_cast<T>
                                   >::type
                                   string_to_number_conversion;
                                      // F is a string and T is signed, float, or unsigned


        typedef typename EvalTypeIf<
                                            inputIsNumber,      // and thus the output is a string
                                            string_to_number_conversion,
                                            number_to_string_conversion
                                   >::type
                                   standard_conversion;
                                      // We either have string to number or number to string conversion



        typedef typename EvalTypeIf<         (inputIsString && outputIsNumber)
                                          || (inputIsNumber && outputIsString),
                                          You_Forgot_To_Overload_cxxtls_lexical_cast_With_The_Following_Parms<T,F>,
                                          standard_conversion  // supported conversions are limited to string to numbers and vice versa
                                   >::type
                                   conversion_cast_type;
                                      // we either have a conversion we understand or some sort of error.


        // trivial cast or real conversion?
        typedef typename EvalTypeIf< isTrivialConversion,
                                     conversion_cast_type,   // conversion required
                                     trivial_cast<T, F>      // simple assignment style conversion
                                   >::type 
                                   type;
                                      // this type will ever let us call it's cast function and get
                                      // the correct coersion, conversion, or an link error.
    };

    template<class To, class From>  To lexical_cast(From v)
       /// Implement a trivial cast or complex conversion from type From to type To.
       /// And when doing so, use a standard conversion format for numbers to strings
       /// and vice versa.
    {
        typedef typename lexical_cast_helper<To, From>::type  ExecutionType;

        To tmp;
        
        ExecutionType::castFunc(tmp, v);

        return tmp;

    }
       ///<
       ///  The generalized template signature that lets you cast from any "From" type
       ///  to any "To" type.  However, only reasonable casts are possible -- or you
       ///  can manually specialize this class to allow any cast you want -- and take
       ///  the risk that casting makes no sense. 
       ///
       ///  The fact at only reasonable casts are possible means that there must be
       ///  manually specified specializations of this template signature for all
       ///  of the "reasonable" casts.  If you get a link error telling you that some
       ///  some template cast signature is not found, it means that you are doing
       ///  something that is unreasonable -- or at least that it is not currently
       ///  planned for.  You can define your own specialization of this signature,
       ///  or change your code not to use the lexical cast.  See examples in the
       ///  lexical_cast.h file for examples of your own specializations.
       ///
       ///  Remember that cast should be producing a reversible operation:  if you 
       ///  cast from A to B, you should be able to cast from B to A.  Violate this
       ///  rule at your own risk!
       ///
       ///  Note that lexical casts are useful in two ways:  
       ///
       ///  @li On the one hand, you can easly convert from one type to another, 
       ///      such as int's to strings.
       ///
       ///  @li  On the other hand, the fact that the signature of the lexical cast
       ///       template lets you compute the From and the To types based on
       ///       the parameters to some calling template, allows you to automate
       ///       conversion activities.
       ///
       ///  Here's an example of the second benefit: Suppose you have a information
       ///  derived frome function parameters, you can then cast from one type of data
       ///  in the function parameters to some other data type from the function parameters:
       ///
       ///@begincode
       ///
       ///  template<class A, class B>
       ///  void func(A a, B b)
       ///  {
       ///      A tmp = cxxtls::lexical_cast<A>(b);
       ///  }
       ///
       ///@endcode
       ///
       ///  And, let's not forget the fact that class member pointer objects have two pieces
       ///  of type information to be gleaned by the compiler:  the class type and the object
       ///  type:
       ///
       ///@begincode
       ///
       ///  template<class ObjectType, class ClassName>
       ///  std::string ConvertAnyMemberToString(ObjectType ClassName::* p, ClassName *someObject)
       ///  {
       ///     return cxxtls::lexical_cast<std::string>( someObject->::* p);
       ///  }
       ///
       ///@endcode
       ///
       ///  The above function lets you convert any class member to a string, so long as the
       ///  conversion makes sense.  The reverse conversion also works:
       ///
       ///@begincode
       ///
       ///  template<class ObjectType, class ClassName>
       ///  void ConvertAnyMemberFromString(ObjectType ClassName::* p, ClassName *someObject, std::string const &input)
       ///  {
       ///     someObject->::* p = cxxtls::lexical_cast<ObjectType>(input);
       ///  }
       ///
       ///@endcode


    template<class To, class From>  To lexical_cast(From v, fmtspec s)
       /// perform a simple cast or a formatted conversion from From type to the To type.
       /// the format is defined by class @ref cxxtls::fmtspec.
    {
        typedef typename lexical_cast_helper<To, From>::type  ExecutionType;

        To tmp;
        
        ExecutionType::castFunc(tmp, v, s);

        return tmp;

    }


    template<>
    inline std::string lexical_cast<std::string>(char c)
        /// cast from a single signed character to a string.
        ///
        /// This specialization is needed because the default lexical_cast<> invocation
        /// with treat c as an 8 bit signed integer not as a 'character'.
    {
       std::string rv;

       rv += c;

       return rv;
    }

    template<>
    inline char lexical_cast<char>(std::string s)
       /// cast from a string to a single signed character.
       /// This only works if the string's length is exactly 1
       /// otherwise you get a bad cast exception.
       ///
       /// This specialization is necessary because the default
       /// specialization treats <char> as signed 8 bit numbers,
       /// not as characters.
    {
       if(s.empty() || s.size() > 1)
          throw bad_lexical_cast<char,std::string>();

       return s[0];
    }

//===============================================================================================================


    template<>
    inline std::string lexical_cast<std::string>(char c, fmtspec s)
        /// cast from a single character to a string obeying a
        /// format specification
    {
       std::string rv;

       if(s.width < 0)
       {
          int w = 0 - s.width;

          --w; 

          rv += c;

          while(w--)
             rv += ' ';

       }
       else
       {
          if(s.width)
          {
            int w = s.width-1;

            while(w--)
              rv += ' ';
          }

          rv += c;

       }



       return rv;
    }


}




#endif
