#include <cxxtls/foreach.h>

#include <cmath>
#include <iostream>
#include <vector>
#include <map>
#include <set>

//
// Copyright 2013,2016, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.


///@file
///
///  This file defines functions that are used to analyze time series.
///
///  @see arithmeticMean
///
///  @see autoRegressionFunction


namespace cxxtls
{

template<class Iterator>
long double minimum(Iterator first, Iterator last)
{
    if(first == last)
       return 0; // technically, this should be an exception, I guess

    long double tmp = *first++;

    while(first != last)
    {
       long double current = *first++;
       
       if(current < tmp)
          tmp = current;
    }

    return tmp;
}

template<class Container>
long double minimum(Container const &c)
   /// @return the mean of the numbers in a container.  The mean
   /// is the sum of the numbers divided by the count
{
    return minimum(c.begin(), c.end());
}

template<class Iterator>
long double maximum(Iterator first, Iterator last)
{
    if(first == last)
       return 0; // technically, this should be an exception, I guess

    long double tmp = *first++;

    while(first != last)
    {
       long double current = *first++;
       
       if(current > tmp)
          tmp = current;
    }

    return tmp;
}

template<class Container>
long double maximum(Container const &c)
   /// @return the mean of the numbers in a container.  The mean
   /// is the sum of the numbers divided by the count
{
    return maximum(c.begin(), c.end());
}


template<class Iterator>
long double arithmeticMean(Iterator first, Iterator last, long double &minimum, long double &maximum, size_t &count)
   /// @return the mean, the minimum, and the maximum of a sequence of numbers.   The mean
   /// is the sum of the numbers divided by the count thereof.
{
   long double tmp=0;

   if(first != last)
   {
      minimum = maximum = *first;
   }

   count = 0;

   while(first != last)
   {
      long double cur = *first++;

      if(cur < minimum)
         minimum = cur;
      else
      if(cur > maximum)
         maximum = cur;

      tmp += cur;
      ++count;
   }

   tmp /= count;

   return tmp;

}

template<class Iterator>
long double arithmeticMean(Iterator first, Iterator last)
   /// @return the mean of a sequence of numbers.   The mean
   /// is the sum of the numbers divided by the count thereof.
{
    long double minimum = 0;
    long double maximum = 0;
    size_t      count   = 0;

    return arithmeticMean(first, last, minimum, maximum, count);

}


template<class Container>
long double arithmeticMean(Container const &c)
   /// @return the mean of the numbers in a container.  The mean
   /// is the sum of the numbers divided by the count
{
    long double minimum = 0;
    long double maximum = 0;
    size_t      count   = 0;

    return arithmeticMean(c.begin(), c.end(), minimum, maximum, count);
}


template<class Iterator>
long double standardDeviation(Iterator first, Iterator last, long double &mean, long double &minimum, long double &maximum, size_t &count)
    /// @return the standard deviation, mean, min, max, and count of numbers found in a container.   The container is defined by a range of iterators.
{
   if(first == last)
   {
      count = 0;
      return 0;
   }

   mean = arithmeticMean(first, last, minimum, maximum, count);

   long double rv = 0;

   while(first != last)
   {
       long double tmp = *first++ - mean;

       rv += (tmp * tmp);
   }

   rv = sqrt(rv / count);

   return rv;


} // standardDeviation

template<class Container>
long double standardDeviation(Container const &c)
    /// @return the standard deviation of numbers found in a container.
{
   long double mean    = 9;
   long double minimum = 0;
   long double maximum = 0;
   size_t      count   = 0;

   return standardDeviation(c.begin(), c.end(), mean, minimum, maximum, count);


} // standardDeviation


template<class Container>
long double standardDeviation(Container const &c, long double &mean)
    /// @return the standard deviation of numbers found in a container.
{
   long double minimum = 0;
   long double maximum = 0;
   size_t      count   = 0;

   return standardDeviation(c.begin(), c.end(), mean, minimum, maximum, count);


} // standardDeviation


template<class Container>
long double autoCorrelationFunction(Container const &c, size_t lag=1)
///  @return the auto-correlation function value for the time series stored in a 
///  container using a given time lag value.
///
///  This function is used to search within a time series sequence to see if you have
///  an approximately repeated sub-sequence.  By approximately, it is meant
///  that if you have noisy data, the noise will be rounded out by the
///  algorithm so that you can detect the approximate numerical sequence contained.
///
///  The search is performed by repeatedly calling the function with lag values running from 0 to
///  c.size() -1.  You look for return values that are 1 or close to it as indicators that the current
///  lag value marks the beginning of a new repetition of the subsequence that runs from location 0 -- to the
///  current lag value.  The current lag value is then the length of the repeating subsequence.
///
///  For a lag of 0, the autoCorrelationFunction() function should return 1.  If you have an EXACTLY
///  repeated sub-sequence, of say 3 numbers wide, then every fourth number in the container will begin
///  a new repetition of the sequence.  For lag values of 0, 3, 6, 9, the autoCorrelationFunction() result will be 1.
///  It should not be 1 for any other locations unless you actually have a repeated sub-sequence that you 
///  don't recognize.
///
///  In real world though, noise interferes with exact matches.  This function smooths
///  out these minor numerical hiccups -- but the cost of this is that the periods are not marked by an
///  acf (autocorrlation function) value of 1.0.  Instead, you must look for approximately 1, or perhaps even
///  a smaller  if the data is very noisy.  
///
///  0.95 is a commonly used "close enough" value for the acf to declare a sequence length.  However, the number
///  may be lower depending on the data noise.  I did an experiment with the hourly temperature readings at 
///  Philadelphia airport over a 4 day period.  I did not find acf > 0.95 every 24 hours like you would expect, 
///  but I did get obvious maxima in the acf values at about 24 hours.  The acf fell as low as .875 around some
///  time periods, but still you could clearly see the rise and fall of the correlation
///  value in a 24 hour period.  Note that you can't pick .875 as a generalized acceptable value because there
///  are many sequences where the actual acf value is actually 1.  Basically you have to explore the values
///  of the acf function over your sequence to understand the range so as to decide what the cutoff will be.
///
///  Note that for lag of 0, the acf will always be 1 -- assuming you have at least 1 item in the input series.
///  The full subsequence obviously counts as a "repeat".  But you are typically looking for subsequences.   If
///  you don't find any other local maxima in the acf values (across all lags) then there are no sub-sequences.
///  Note that noisy data can result in their being a trailing 1 near the end of the input sequence.  Care must
///  be taken to understand this effect in real world data cases.
///
///  If your first acf "period identifier value" after location zero in the time series 0 is near the end of the 
///  sequence, then you don't have  repetititions within the sequence.
///
///  I have also noticed that when the range of numbers is compressed, extra sub-sequence correlations seem to appear.
///  I'm not sure why.  I have also noticed that positive and negative swings away from the mean of the signal are
///  treated the same.  So I suggest that you bias the data  into a strictly positive range before feeding it to 
///  this function.
///
{
   if(c.size() < 2)
      return 1.0;

   if(lag >= c.size() -1)
      return 0;

   typename Container::const_iterator first = c.begin(), last = c.end();

   typename Container::const_iterator second = first;  

   typename Container::const_iterator periLast = last; 

   while(lag--)
   {
      ++first;
      --last;
   }


    long double meanI1, meanI2;

    meanI1 = arithmeticMean(first, periLast);  // 0->(n-1)
    meanI2 = arithmeticMean(second, last);     // 1->n

    // note:  there are two ways of computing this function's value:
    //        1).  for small sequences a more complex algorithm must be
    //             used to guarantee no information is loss.
    //        2).  A simpler algorithm that only for long sequences where
    //             a tiny bit of round off is preferable to the added
    //             cost of extra calculations.
    //
    //        this function provides only option one.

    long double numerator = 0;
    long double denominator1 = 0;
    long double denominator2 = 0;

    while(first != periLast)               // range:  1->(N-1)
    {
        long double l = *first++;          // c[n]
        long double r = *second++;         // c[n+1]

        long double lvariance = (l - meanI1);
        long double rvariance = (r - meanI2);

        long double tmp = lvariance * rvariance;

        denominator1 += lvariance * lvariance;
        denominator2 += rvariance * rvariance;

        numerator += tmp;


    }

    denominator1 = sqrt(denominator1);
    denominator2 = sqrt(denominator2);

    long double tmp = denominator1 * denominator2;

    if(tmp == 0)
       return 0;

    tmp = numerator / tmp;

    if(tmp < 0)
       tmp = 0 - tmp;

    return tmp;


} // autoCorrelationFunction


struct TimeSeriesAnalysis
   ///  This class represents all all the information that can be efficiently
   ///  calculated about a time series using the tools above.  This class
   ///  exists to simplify the process of analysis of a time series.
   ///
   ///  You construct an object of this class to get, primarily, the lenghts
   ///  of any subsequences of numbers in a time series.
   ///
   ///  Note that many functions in this class are static so that they can be
   ///  used on data outside the class if need be.
   ///
   ///  After construction, all the data is finalized.
   ///
   ///  the highest level "output" of the construction process is:
   ///
   ///  @li acfPeriodWidths_ -- it contains a list of subsequence periods detected.
   ///  @li acfPeriods_ -- locations where some sort of periodicity, main sequence
   ///                     or subsequence, were detected.   The .first members is a
   ///                     relative measure of the largeness of the period.  If the
   ///                     .first member is 1, it is clearly a top level sequence,
   ///                     with subsequences holding smaller .first members.  The 
   ///                     .second member of acfPeriods_[i] is an index into the
   ///                     acf_ array where one repretition of this subsequence might start.
   ///                     
   ///  @li various mean, maxima, minima, and standard deviations.
   ///
{
    std::vector<long double> data_; ///<  A copy of the data in the time series.

    long double minimum_;   ///<  the minimum value of the time series
    long double maximum_;   ///<  the maximum value of the time series
    long double stdDev_;    ///<  the standard deviation of the time series
    long double mean_;      ///<  the mean value of the time series
    size_t      count_;     ///<  the count of points in the container

    std::vector<long double>  acf_;  ///<  the auto-correlation-function values for
                                     ///   the time series.

    long double acfStdDev_;  ///<  the standard deviation of acf_ 
    long double acfMean_;    ///<  mean of the acf_ values

    long double acfLevelStdDev_;  ///<  the standard deviation of acfLevelValues_
    long double acfLevelMean_;    ///<  the mean of acfLevelValues_

    typedef std::map<long double, size_t> ACF_LEVELS_TYPE;  ///<  The locations within acf_ of local maxima
                                                            ///   of the acf_ function.

    ACF_LEVELS_TYPE           acfPeriods_;       ///<  Locations within acf_ where various acf maxima occur.
    std::vector<long double>  acfLevelValues_;   ///<  An ordered list of the various acf maxima.

    std::set<size_t>          acfPeriodWidths_;  ///<  The sizes of REASONABLE periods found in acf_.  That is,
                                                 ///   the periods that are not ridiculously small  and those
                                                 ///   that are not ridiculously large.

    std::set<size_t>          acfMainSequenceTimeIndices_;   ///<  indices into acf where the main sequence period
                                                             ///   begins and repeats
    

    template<class Iterator>
    TimeSeriesAnalysis(Iterator first, Iterator last)
        ///  Construct the TimeSeriesAnalysis object and perform all
        ///  needed computations
    {
        populate(first,last);
    }

    template<class Container>
    TimeSeriesAnalysis(Container const &c)
        ///  Construct the TimeSeriesAnalysis object and perform all
        ///  needed computations
    {

       populate(c.begin(), c.end()); 
    }

    template<class Iterator>
    void populate(Iterator first, Iterator last)
        ///  Perform the complete set of computations that populate all members
        ///  of *this.
    {
       data_.assign(first, last);

        stdDev_ = standardDeviation(first, last, mean_, minimum_, maximum_, count_);

        // the follow code is REALLY slow

        for(size_t lag = 0; lag < count_; ++lag)
            acf_.push_back(autoCorrelationFunction(data_, lag));

        acfStdDev_ = standardDeviation(acf_, acfMean_);
            
        discoverAcfPeriods(acf_, acfPeriods_, acfLevelValues_);   
                                                 // find subsequences within the acf_ data where
                                                 // the sequence starting at lag=0 is repeated,
                                                 // or find other major types of subsequences -- the two
                                                 // can't easily be disambiguated.

        acfLevelStdDev_ = standardDeviation(acfLevelValues_, acfLevelMean_);


        std::pair<size_t,size_t> rv = discoverMainSeriesPeriodWidths(acfPeriods_, acfPeriodWidths_, acfLevelStdDev_, acfMainSequenceTimeIndices_);  // "acf level std dev", not just "acf std dev"!

        size_t const &max = rv.first;
        size_t const &min = rv.second;

        if(max < min)
        {
           // this means that the data is bogus

           acfPeriodWidths_.clear();
           acfPeriodWidths_.insert(data_.size());
        }

    }  // populate()

    template<class AcfContainer, class MaximaLocations, class LevelValueOrder>
    static void discoverAcfPeriods(AcfContainer const &acfValues, MaximaLocations &acfLevels, LevelValueOrder &acfLevelValues)
        /// analyze the acf values to determine where sub-periods start within the
        /// acf data.  
    {
       bool goingDown = true;
       long double prev    = acfValues[0];
       long double top     = 1.0;  // at loc 0, we should have a 1.0
       size_t prevI   = 0;

       typedef typename MaximaLocations::value_type  ACF_LEVEL_ENTRY;


       acfLevels.insert(ACF_LEVEL_ENTRY(1.0,0));

       size_t acfCount = acfValues.size();

       for(size_t i = 1; i < acfCount; ++i)    // DOES NOT start with location 0!
                                               // location 0 should be 1.0 if the code
                                               // is correct.
       {
          long double current = acfValues[i];

          if(goingDown)
          {
              if(current < prev)
              {
                 prev = current;
                 continue;
              }
              else
              if(current == prev)
              {
                if(current == top)
                {
                    prevI = i -1;
                    
                    acfLevels.insert( ACF_LEVEL_ENTRY(prev,prevI) );
                    acfLevelValues.push_back(prev);

                    continue;
                }
                else
                {
                   prev = current;
                   continue;
                }
              }
              else
                 goingDown = false; // ie we are now going up
          }
          else
          {
             // going up

             if(current >= prev)
             {
                prev = current;
                continue;
             }
             else
             {
                // prev is a local maximum
                goingDown = true;  // now we are falling
                prevI = i -1;
                top = prev;  // current guy is < prev, so prev was top

                acfLevels.insert( ACF_LEVEL_ENTRY(prev,prevI) );
                acfLevelValues.push_back(prev);
             }

          }
       }

    }  // discoverAcfPeriods


    template<class AcfLevelsMap>
    static 
    std::pair<size_t,size_t> 
    discoverMainSeriesPeriodWidths(AcfLevelsMap const &acfLevels, std::set<size_t> &timeDeltas, long double acfLevelStdDev, std::set<size_t> &timeIndices)
        ///   Discover the main, realistic sequence lengths found in some container described by
        ///   acfLevels.   AcfLevels is a map of acf value to the location within the acf_ container that
        ///   holds the actual data.  AcfLevels contains only instances of local maxima in the acf values.
        ///
        ///   This function returns the set of realistic subsequence lengths and stores it in timeDeltas.
        ///
        ///   @return pair<largest,smallest> subsequence in the table of acf values described by map, acflevels,
        ///           and return a container, timeDeltas, which contains a filtered set of the subsequence
        ///           lengths (referring to the same acf vector as acfLevels).  Note that the largest and smallest
        ///           may not appear in the timeDeltas output vector.  The represent a "statistic" about the data
        ///           process and possibly ignored by this function.
        ///
        ///   @param acfLevels A map of acf values paired with indexes into a vector of said acf values.  The vector
        ///          is assumed to have been created as a sequence of autoCorrelationFunction() call return values
        ///          with the lag parameter running from 0 to time series array size - 1.   The idea is that you take a
        ///          time series, call the autoCorrelationFunction for each member of it.  The first acf result should
        ///          be 1 and others will fall and rise periodically within the acf vector your create.
        ///
        ///   @param timeDeltas An output set of reasonably sized subsequences within the acf vector described by
        ///                     acfLevels and acfStdDev.
        ///
        ///   @param acfLevelStdDev  The standard deviation of the acfLevelValues vecotr
        ///
        ///   @param timeIndices the indices into acf_ where local maxima which are >= than 1- acfLevelStdDev are found
    {


       long double minLevel = 1 - acfLevelStdDev;     // minimum value of a local maxima
                                                       // of interest.

       minLevel -= 1e-7;  // slop in the number when you do a bunch of complex calculations that should
                          // add up to one, it just ain't going to happen.  This number might should be bigger 
                          // ie minLevel might should be reduced even more.  We are basically doing
                          // rough analysis anyway.


       size_t maxDelta = 0;
       size_t minDelta = size_t(-1);  // damn you compiler, you'd better not bitch, you punk!

       {
           
           CXXTLS_FOREACH(typename AcfLevelsMap::value_type const &cur, acfLevels)
           {
              if(cur.first >= minLevel)
              {
                  timeIndices.insert(cur.second);
              }
              
           }

           // timeIndices now tells us where interesting things occur in acf_
           
           size_t largestDelta = 0;
           
           {
               std::set<size_t>::iterator first = timeIndices.begin(), last = timeIndices.end();
           
               while(first != last)
               {
                  size_t cur = *first++;
           
                  if(first != last)
                  {
                     size_t delta = *first - cur;
           

                     if(delta > maxDelta)
                        maxDelta = delta;

                     if(delta < minDelta)
                        minDelta = delta;

                     timeDeltas.insert(delta); 
           
                     if(delta > largestDelta)
                        largestDelta = delta;

                  }
           
           
               }
           }

           size_t meanTimeStdDev = ceil(standardDeviation(timeDeltas));

           // removing ridiculously large time deltas (ie very long sequences)
           // from the return value.
           //
           // We are going to repeatedly scan through the output data looking
           // for individual deltas (ie periods) that are more than one stddev
           // larger than the rest.  In most sequences, stddev should be much smaller
           // than the individual time deltas -- otherwise, what does it mean to have
           // a subsequence -- without repeats of smaller subsequences, what does it mean?
           

           while(largestDelta > meanTimeStdDev)
           {
               // remove items out of range

               std::set<size_t> tmp;    // copy of timeDeltas with big items removed

               largestDelta = 0;      // recompute largest delta as we filter out
                                      // the unusually large items

               CXXTLS_FOREACH(size_t cur, timeDeltas)
               {
                   if(cur < meanTimeStdDev)
                   {

                      if(cur > largestDelta)
                         largestDelta = cur;

                      tmp.insert(cur);
                   }
                   else
                   {
                      // ignore big deltas as being wierd
                   }

               }

               if(tmp.empty())
                  break;

               timeDeltas.swap(tmp);  // swap computed data with old timeDeltas to eliminate wierdos

               // recompute statistics

               meanTimeStdDev = ceil(standardDeviation(timeDeltas));

           }

           // since we break out of the above loop, we might need to recompute
           // largestDelta


           largestDelta = maximum(timeDeltas);

           // now keep only the biggest of the smallest

           {
              std::set<size_t> tmp;

              CXXTLS_FOREACH(size_t cur, timeDeltas)
              {
                 if(cur >= largestDelta - meanTimeStdDev)
                 {
                    tmp.insert(cur);
                 }
              }

              if(tmp.size())
                 timeDeltas.swap(tmp);

           }

       }  // main sequence
    
       return std::pair<size_t,size_t>(maxDelta, minDelta);

    }  // discoverMainSeriesPeriodWidths



};  // TimeSeriesAnalysis


struct insufficient_data : public std::exception {};
struct vertical_line     : public std::exception {};

std::pair<double,double>
linearRegression(std::vector <std::pair<double,double> >const &xy)
/// 
/// @return pair<slope,intercept> given a colleciton of pair<x,y> using
/// linear regression.
///
/// See https://www.easycalculation.com/statistics/learn-regression.php
/// 
/// "Regression Equation(y) = a + bx 
/// Slope(b) = (N?XY - (?X)(?Y)) / (N?X2 - (?X)2) 
/// Intercept(a) = (?Y - b(?X)) / N"
/// 
/// "x and y are the variables. 
/// b = The slope of the regression line 
/// a = The intercept point of the regression line and the y axis. 
/// N = Number of values or elements 
/// X = First Score 
/// Y = Second Score
///  ?XY = Sum of the product of first and Second Scores 
/// ?X = Sum of First Scores 
/// ?Y = Sum of Second Scores 
/// ?X2 = Sum of square First Scores"

{ // linearRegression


   std::vector<double> XtimesY;
   std::vector<double> XtimesX;
   double              sumOfX = 0;
   double              sumOfY = 0;
   double              sumOfXtimesX = 0;
   double              sumOfXtimesY = 0;


   if(xy.size() < 1)
   {
       throw insufficient_data();   
   }

   for(size_t i = 0; i < xy.size(); ++i)
   {
      double x = xy[i].first;
      double y = xy[i].second;

      double xTimesY = x * y;

      XtimesY.push_back(xTimesY);

      sumOfXtimesY += xTimesY;

      double xtimesx = x * x;

      XtimesX.push_back(xtimesx);

      sumOfXtimesX += xtimesx;

      sumOfX += x;
      sumOfY += y;

   }

   // slope numerator = (N?XY - (?X)(?Y))

   double slopeNumerator     = xy.size() * sumOfXtimesY - sumOfX * sumOfY;

   // slope denominator   (N?X2 - (?X)2)

   double slopeDenominator = xy.size() * sumOfXtimesX - sumOfX*sumOfX;

   if(slopeDenominator == 0)
      throw vertical_line();

   double slope = slopeNumerator / slopeDenominator;

// Intercept(a) = (?Y - b(?X)) / N"

   double interceptNumerator = sumOfY - slope * sumOfX;

   double intercept = interceptNumerator / xy.size();

   return std::pair<double,double>(slope,intercept);



}  // linearRegression


} // namespace cxxtls
